using System;
using System.Collections.Generic;
using System.Linq;
using Urho;
using Urho.Audio;
using Urho.Gui;
using Urho.Physics;
using Urho.Resources;
using Urho.Urho2D;

namespace UrhoSharpGoogleTango.Scripts
{
    public class DemoScene : Application
    {
        private readonly List<MonsterComponent> _activeMonsters = new List<MonsterComponent>();
        private readonly List<BulletComponent> _bullets = new List<BulletComponent>();
        private GamePhase _currentPhase;
        private int _currentStage;
        private int _currentWave;

        private GameAssets _gameAssets;
        private StaticModel _luasModel;
        private float _phaseCooldown;
        private int _phaseCounter;
        private float _helpCooldown;

        private SoundSource _soundSource;
        private bool _spawnABulletOnNextFrame;
        private Stages _stages;
        private ResourceCache cache;

        private UrhoConsole console;
        private DebugHud debugHud;
        private Sprite logoSprite;
        private readonly Random rnd = new Random();
        /**/
        private Scene scene;
        private UI ui;
        private SoundSource _soundSource2;
        private StaticModel _worldModel;
        private SoundSource _soundSource3;
        //[Preserve]
        public DemoScene(ApplicationOptions options) : base(options)
        {
        }

        protected Node CameraNode { get; set; }
        protected MonoDebugHud MonoDebugHud { get; set; }


        public Texture2D VideoTexture { get; set; }

        public int VideoTextureId
        {
            get { return -1; }
        }

        protected Vector3[] PointCouldPoints { get; set; }

        public event EventHandler OnSceneCreated;
        public event EventHandler OnEndViewRender;
        public event EventHandler OnBeginViewRender;

        private void HandleEndViewRender(EndViewRenderEventArgs obj)
        {
            OnEndViewRender?.Invoke(this, EventArgs.Empty);
        }

        private void HandleBeginViewRender(BeginViewRenderEventArgs obj)
        {
            OnBeginViewRender?.Invoke(this, EventArgs.Empty);
        }

        protected override void Start()
        {
            base.Start();
            CreateScene();
            ResetGame();


            Input.Enabled = true;
            MonoDebugHud = new MonoDebugHud(this);
            MonoDebugHud.Show();

            CreateConsoleAndDebugHud();
            Graphics.FlushGPU = true;

            SetupViewport();

            Renderer.BeginViewRender += HandleBeginViewRender;
            Renderer.EndViewRender += HandleEndViewRender;

            //Input.JoystickConnected += HandleHandleJoystickButtonDown;
            Input.JoystickButtonDown += HandleJoystickButtonDown;
            //Input.KeyDown += HandleKeyDown;
            Input.JoystickConnected += HandleJoystickConnected;
            Input.JoystickDisconnected += HandleJoystickDisconnected;
        }

        private void ResetGame()
        {
            foreach (var monsterComponent in _activeMonsters)
            {
                monsterComponent.Node.Remove();
            }
            foreach (var monsterComponent in _bullets)
            {
                monsterComponent.Node.Remove();
            }

            _luasModel.Model = _gameAssets.LuasAnimation.Last();
            _currentPhase = GamePhase.Traveling;
            _currentStage = 0;
            _currentWave = 0;
            _phaseCooldown = 0;
        }

        protected Text SimpleCreateInstructions(string text = "")
        {
            var textElement = new Text
            {
                Value = text,
                HorizontalAlignment = HorizontalAlignment.Center,
                VerticalAlignment = VerticalAlignment.Center
            };
            textElement.SetFont(ResourceCache.GetFont("Fonts/Anonymous Pro.ttf"), 15);
            UI.Root.AddChild(textElement);
            return textElement;
        }


        private void HandleJoystickDisconnected(JoystickDisconnectedEventArgs obj)
        {
        }

        private void HandleJoystickConnected(JoystickConnectedEventArgs obj)
        {
        }

        private void HandleKeyDown(KeyDownEventArgs obj)
        {
            _spawnABulletOnNextFrame = true;
        }

        private void HandleHandleJoystickButtonDown(JoystickConnectedEventArgs obj)
        {
        }

        private void HandleJoystickButtonDown(JoystickButtonDownEventArgs obj)
        {
            _spawnABulletOnNextFrame = true;
        }


        private void SetupViewport()
        {
            var graphics = Graphics;
            var renderer = Renderer;
            var cache = ResourceCache;

            var viewport = new Viewport(Context, scene, CameraNode.GetComponent<Camera>(), null);

            renderer.SetViewport(0, viewport);
        }

        protected override void OnUpdate(float timeStep)
        {
            base.OnUpdate(timeStep);
        }

        private void CreateScene()
        {
            var haveShadow = false;

            cache = ResourceCache;
            _gameAssets = new GameAssets(cache);

            _stages = new Stages(_gameAssets, cache);

            scene = new Scene(Context);
            scene.CreateComponent<Octree>();
            var physicsWorld = scene.CreateComponent<PhysicsWorld>();
            physicsWorld.SubscribeToPhysicsPreStep(HandlePhysicsPreStep);

            {
                var node = scene.CreateChild();
                node.Position = new Vector3(-32, -2, -32);
                _worldModel = node.CreateComponent<StaticModel>();
                _worldModel.Model = cache.GetModel("Models/castle.mdl");
                _worldModel.SetMaterial(_gameAssets.DefaultMaterial);
            }
            {
                var scale = 4.0f/16.0f;
                var node = scene.CreateChild();
                node.Position = new Vector3(-8, -1, -16)*scale;
                node.Scale = new Vector3(scale, scale, scale);
                _luasModel = node.CreateComponent<StaticModel>();
                _luasModel.Model = _gameAssets.LuasAnimation.Last();
                _luasModel.SetMaterial(_gameAssets.DefaultMaterial);
            }
            //SpawnWaveOfMonsters(scene, _deer, 5);

            //scene.CreateComponent<Octree>();
            // Create octree, use default volume (-1000, -1000, -1000) to (1000, 1000, 1000)
            //scene.CreateComponent<Octree>();

            // Create a Zone component for ambient lighting & fog control
            var zoneNode = scene.CreateChild("Zone");
            var zone = zoneNode.CreateComponent<Zone>();
            zone.SetBoundingBox(new BoundingBox(-1000.0f, 1000.0f));
            zone.AmbientColor = new Color(0.5f, 0.5f, 0.5f);
            //zone.FogColor = new Color(1.0f, 1.0f, 1.0f);
            //zone.FogStart = 500.0f;
            //zone.FogEnd = 750.0f;


            // Create skybox. The Skybox component is used like StaticModel, but it will be always located at the camera, giving the
            // illusion of the box planes being far away. Use just the ordinary Box model and a suitable material, whose shader will
            // generate the necessary 3D texture coordinates for cube mapping
            //var skyNode = scene.CreateChild("Sky");
            //skyNode.SetScale(500.0f); // The scale actually does not matter
            //var skybox = skyNode.CreateComponent<Skybox>();
            ////skybox.Model = cache.GetModel("Models/sky01.mdl");
            ////skybox.SetMaterial(cache.GetMaterial("Materials/Sky.xml"));
            //skybox.Model = cache.GetModel("Models/Box.mdl");
            //skybox.SetMaterial(cache.GetMaterial("Materials/Skybox.xml"));


            // Create the camera. Limit far clip distance to match the fog
            CameraNode = new Node();
            var camera = CameraNode.CreateComponent<Camera>();
            camera.FarClip = 750.0f;

            _soundSource = CameraNode.CreateComponent<SoundSource>();
            _soundSource.Gain = 2.0f;

            _soundSource2 = CameraNode.CreateChild("second sound").CreateComponent<SoundSource>();
            _soundSource2.Gain = 2.0f;

            _soundSource3 = CameraNode.CreateChild("shot sound").CreateComponent<SoundSource>();
            _soundSource3.Gain = 2.0f;
            // Set an initial position for the camera scene node above the plane
            //SlideCamera(0);

            SetCameraPositoin(new Vector3(0, 0, 0));

            OnSceneCreated?.Invoke(this, EventArgs.Empty);
        }

        private void HandlePhysicsPreStep(PhysicsPreStepEventArgs obj)
        {

            if (_spawnABulletOnNextFrame)
            {
                _spawnABulletOnNextFrame = false;
                SpawnBullet();
            }
            foreach (var monster in _activeMonsters)
            {
                monster.FixedUpdate(obj.TimeStep);
            }
            for (var index = _bullets.Count - 1; index >= 0; index--)
            {
                var bulletComponent = _bullets[index];
                bulletComponent.FixedUpdate(obj.TimeStep);

                for (var monsterIndex = _activeMonsters.Count - 1; monsterIndex >= 0; monsterIndex--)
                {
                    var monster = _activeMonsters[monsterIndex];
                    if (bulletComponent.IsHit(monster))
                    {
                        monster.Health -= 1;
                        if (monster.Health <= 0)
                        {
                            monster.Node.RemoveComponent<StaticModel>();
                            monster.Explode(_gameAssets);
                            //var particleEmitter = monster.Node.CreateComponent<ParticleEmitter2D>();
                            ////explodeNode.SetScale(1.5f);
                            //particleEmitter.Effect = _gameAssets.ExplosionEffect;
                            ////scene.RemoveChild(monster.Node);
                            _activeMonsters.RemoveAt(monsterIndex);
                        }

                        scene.RemoveChild(bulletComponent.Node);
                        _bullets.RemoveAt(index);
                        goto nextiteration;
                    }
                }

                if (bulletComponent.Node.Position.LengthSquared > 32*32)
                {
                    scene.RemoveChild(bulletComponent.Node);
                    _bullets.RemoveAt(index);
                }
                nextiteration:
                ;
            }

            HandlePhase(obj.TimeStep);
        }

        private void HandlePhase(float timeStep)
        {
            switch (_currentPhase)
            {
                case GamePhase.Shooting:
                    HandleShootingPhase(timeStep);
                    break;
                case GamePhase.ClosingDoors:
                    HandleClosingDoors(timeStep);
                    break;
                case GamePhase.OpeningDoors:
                    HandleOpeningDoors(timeStep);
                    break;
                case GamePhase.Traveling:
                    HandleTraveling(timeStep);
                    break;
            }
        }
        private void HandleTraveling(float timeStep)
        {
            _phaseCooldown += timeStep;
            if (_phaseCooldown < 7.0f)
                return;
            _phaseCooldown = 0;
            _phaseCounter = 0;
            _currentPhase = GamePhase.OpeningDoors;
            _worldModel.Model = CurrentStage.Model;
            _worldModel.Node.Position = CurrentStage.ModelPosition;
            _worldModel.SetMaterial(_gameAssets.DefaultMaterial);
            _soundSource2.Play(_gameAssets.DoorSound);
        }
        private void HandleClosingDoors(float timeStep)
        {
            _phaseCooldown += timeStep;
            if (_phaseCooldown < 0.14f)
                return;
            _phaseCooldown = 0;
            ++_phaseCounter;
            if (_phaseCounter >= 4)
            {
                _phaseCooldown = 0;
                _phaseCounter = 0;
                _currentPhase = GamePhase.Traveling;
                if (CurrentStage.Anouncement != null)
                _soundSource.Play(CurrentStage.Anouncement);
                return;
            }
            
            _luasModel.Model = _gameAssets.LuasAnimation[_phaseCounter];
            _luasModel.SetMaterial(_gameAssets.DefaultMaterial);
        }
        private void HandleOpeningDoors(float timeStep)
        {
            _phaseCooldown += timeStep;
            if (_phaseCooldown < 0.1f)
                return;
            _phaseCooldown = 0;
            ++_phaseCounter;
            if (_phaseCounter >= 4)
            {
                _phaseCooldown = 0;
                _phaseCounter = 0;
                _currentPhase = GamePhase.Shooting;
                return;
            }

            _luasModel.Model = _gameAssets.LuasAnimation[3-_phaseCounter];
            _luasModel.SetMaterial(_gameAssets.DefaultMaterial);
        }
        public StageDescription CurrentStage
        {
            get
            {
                return _stages.List[_currentStage];
            }
        }
        public WaveDescription CurrentWave
        {
            get
            {
                if (_currentWave >= CurrentStage.Waves.Length)
                    return null;
                return CurrentStage.Waves[_currentWave];
            }
        }
        private void HandleShootingPhase(float timeStep)
        {
            if (_activeMonsters.Count > 0)
            {
                _helpCooldown += timeStep;
                if (_helpCooldown > 4)
                {
                    var minD = float.MaxValue;
                    Sound sound = null;

                    var dirs = new []
                    {
                        Tuple.Create(MonsterComponent.Transform(Vector3.Forward, CameraNode.Rotation), (Sound)null),
                        Tuple.Create(MonsterComponent.Transform(Vector3.Left, CameraNode.Rotation), (Sound)_gameAssets.Left),
                        Tuple.Create(MonsterComponent.Transform(Vector3.Right, CameraNode.Rotation), (Sound)_gameAssets.Right),
                        Tuple.Create(MonsterComponent.Transform(Vector3.Back, CameraNode.Rotation), (Sound)_gameAssets.Back),
                    };

                    foreach (var monster in _activeMonsters)
                    {
                        var dir = CameraNode.Position- monster.Node.Position;

                        Sound side = null;
                        var minProj = float.MaxValue;
                        foreach (var tuple in dirs)
                        {
                            var proj = Vector3.Dot(tuple.Item1, dir);
                            if (proj < minProj)
                            {
                                minProj = proj;
                                side = tuple.Item2;
                            }
                        }
                        if (side != null)
                        {
                            var d = dir.LengthSquared;
                            if (d < minD)
                            {
                                minD = d;
                                sound = side;
                            }
                        }
                    }
                    _helpCooldown = 0;
                    if (sound != null)
                        _soundSource2.Play(sound);
                }
            }
            var wave = CurrentWave;
            if (wave == null)
            {
                if (_activeMonsters.Count > 0)
                    return;
                _helpCooldown = 0;
                _currentWave = 0;
                _soundSource2.Play(_gameAssets.DoorClosingSound);
                _currentPhase = GamePhase.ClosingDoors;
                _currentStage = (_currentStage + 1)% _stages.List.Length;
                _phaseCooldown = 0;
                _phaseCounter = 0;
                return;
            }
            _phaseCooldown += timeStep;
            if (_phaseCooldown >= wave.WaveDelay)
            {
                SpawnWaveOfMonsters(scene, CurrentWave);
                ++_currentWave;
                _phaseCooldown = 0;
                _phaseCounter = 0;
                return;
            }
        }

        private void SpawnWaveOfMonsters(Scene scene, WaveDescription wave)
        {
            for (var i = 0; i < wave.Count; ++i)
            {
                var a = rnd.NextDouble()*(wave.MaxAngle-wave.MinAngle)+ wave.MinAngle;
                var scale = (float)(rnd.NextDouble() * (wave.MaxScale - wave.MinScale) + wave.MinScale);
                var dist = (wave.Monster.StartDistance + (rnd.NextDouble() - 0.5f) * wave.Monster.StartDistanceD);
                var x = (float) (Math.Cos(a)* dist);
                var y = (float)(Math.Sin(a) * dist);
                SpawnAMonster(scene, wave.Monster, new Vector3(x, 0, y)).Scale = new Vector3(scale, scale, scale);
            }
        }

        private Node SpawnAMonster(Scene scene, MonsterData deer, Vector3 pos)
        {
            var node = scene.CreateChild();
            node.Position = pos;
            var d = -pos;
            d.Normalize();
            if (Vector3.Dot(deer.Direction, d) < 0.99f)
            {
                node.Rotation = Quaternion.FromRotationTo(deer.Direction, d);
            }
            var model = node.CreateComponent<StaticModel>();
            var monsterComponent = new MonsterComponent(deer);
            node.AddComponent(monsterComponent);
            _activeMonsters.Add(monsterComponent);
            return node;
        }

        private void SpawnBullet()
        {
            var node = scene.CreateChild();
            node.Position = CameraNode.Position +
                            MonsterComponent.Transform(new Vector3(0, -0.1f, 0), CameraNode.Rotation);
            node.Rotation = CameraNode.Rotation;
            var model = node.CreateComponent<StaticModel>();
            model.Model = _gameAssets.BulletModel;
            model.SetMaterial(_gameAssets.DefaultMaterial);
            var monsterComponent = new BulletComponent();
            node.AddComponent(monsterComponent);
            _bullets.Add(monsterComponent);
            _soundSource3.Play(_gameAssets.Shot);
        }

        public void SetCameraPositoin(Vector3 vector3)
        {
            try
            {
                if (CameraNode != null)
                {
                    CameraNode.Position = vector3 + new Vector3(0, 1.9f, 0);
                }
            }
            catch (InvalidOperationException)
            {
            }
        }

        public void SetCameraRotation(Quaternion quaternion)
        {
            try
            {
                if (CameraNode != null)
                    CameraNode.Rotation = quaternion;
            }
            catch (InvalidOperationException)
            {
            }
        }


        private void CreateConsoleAndDebugHud()
        {
            //var xml = cache.GetXmlFile("UI/DefaultStyle.xml");
            //console = Engine.CreateConsole();
            //console.DefaultStyle = xml;
            //console.Background.Opacity = 0.8f;

            debugHud = Engine.CreateDebugHud();
            //debugHud.DefaultStyle = xml;
        }

        public void OnKeyDown()
        {
            _spawnABulletOnNextFrame = true;
        }
    }
}